import subprocess
import sys
from pathlib import Path

# Make sure not to import rich here, since it's an optional dependency
# Some tests assert behaviour that's predicated on it not yet being imported
import pytest


@pytest.mark.parametrize(
    "args", [[sys.executable, "-m", "typeshed_stats"], ["typeshed-stats"]]
)
def test_running_from_command_line(complete_typeshed: Path, args: list[str]) -> None:
    result = subprocess.run(
        [*args, "--typeshed-dir", str(complete_typeshed)], check=False
    )
    code = result.returncode
    assert code == 0
    result2 = subprocess.run([*args, "--help"], check=False)
    code = result2.returncode
    assert code == 0



"""Library and command-line tool for getting stats on various typeshed packages."""

import sys

if sys.version_info < (3, 11):  # noqa: UP036
    raise ImportError("Python 3.11+ is required!")

__all__: list[str] = []

if __name__ == "__main__":
    from ._cli import main

    main()



import random
import string
import sys
import textwrap
from collections.abc import Callable, Iterator, Sequence
from pathlib import Path

# Make sure not to import rich or markdown here, since they're optional dependencies.
# Some tests assert behaviour that's predicated on these modules not yet being imported.
import pytest

from typeshed_stats.gather import (
    AnnotationStats,
    CompletenessLevel,
    FileInfo,
    PackageInfo,
    PackageStatus,
    PyrightSetting,
    StubtestSettings,
    StubtestStrictness,
    UploadStatus,
)


@pytest.fixture(autouse=True)
def ensure_optional_dependencies_not_imported() -> Iterator[None]:
    previously_imported_modules = set(sys.modules)
    yield
    newly_imported_thirdparty_modules = {
        module_name
        for module_name in sys.modules
        if (
            module_name not in previously_imported_modules
            and module_name not in sys.stdlib_module_names
        )
    }
    for module_name in newly_imported_thirdparty_modules:
        del sys.modules[module_name]


def random_package_name() -> str:
    def random_string() -> str:
        return "".join(
            random.choice(string.ascii_letters) for _ in range(random.randint(1, 10))
        )

    while (result := random_string()) in {"stdlib", "gdb"}:
        continue  # pragma: no cover
    return result


def write_metadata_text(typeshed: Path, package_name: str, data: str) -> None:
    metadata = typeshed / "stubs" / package_name / "METADATA.toml"
    metadata.write_text(data, encoding="utf-8")


@pytest.fixture(scope="session")
def EXAMPLE_PACKAGE_NAME() -> str:
    return random_package_name()


@pytest.fixture
def typeshed(EXAMPLE_PACKAGE_NAME: str, tmp_path: Path) -> Path:
    typeshed = tmp_path
    (typeshed / "stdlib").mkdir()
    stubs_dir = typeshed / "stubs"
    stubs_dir.mkdir()
    (stubs_dir / EXAMPLE_PACKAGE_NAME).mkdir()
    return typeshed


@pytest.fixture(scope="session")
def real_typeshed_package_names() -> frozenset[str]:
    return frozenset({"emoji", "protobuf", "appdirs"})


@pytest.fixture(scope="session")
def example_stub_source() -> str:
    return textwrap.dedent(
        """
        import _typeshed
        import builtins
        import email
        import typing
        import typing_extensions
        from _typeshed import Incomplete
        from collections.abc import Iterable
        from types import GenericAlias
        from typing import Any, TypeAlias

        a: int
        b: str = ...
        c: Any | None
        d: Any
        d: Incomplete
        e: Iterable[typing.Any]
        f: _typeshed.Incomplete
        g: _typeshed.StrPath
        h: email.message.Message

        # TypeAliases should not count towards the number of "annotated variables"
        i: TypeAlias = int
        j: typing.TypeAlias = str
        k: typing_extensions.TypeAlias = bytes

        class Spam:
            a: tuple[typing.Any, ...] | None
            b = ...
            c: int = ...
            d: typing.Sized
            def __init__(__self) -> None: ...
            def __class_getitem__(cls, item: Any) -> GenericAlias: ...

            # This shouldn't count as an unannotated parameter --
            # it's just "self" with a weird name
            def weird_named_first_param(x): ...

        def func1(arg): ...
        def func2(arg: int): ...
        def func3(arg: Incomplete | None = ...): ...
        def func4(arg: Any) -> Any: ...
        def func5(*args: int, **kwargs: str) -> bool: ...
        def func6(*args, **kwargs): ...
        def func7(*, arg): ...
        def func8(*, arg1: int, arg2: str): ...

        class Eggs:
            def __new__(cls) -> Eggs: ...
            def __init__(self) -> None: ...
            def __init_subclass__(cls) -> None: ...
            async def func5(self, arg): ...
            @staticmethod
            async def func6(arg: str) -> list[bytes]: ...
            @builtins.staticmethod
            def func6_pt_5(arg: str) -> None: ...
            def func7(self, arg: Any) -> _typeshed.Incomplete: ...
            @classmethod
            def class_method(cls, eggs: Incomplete): ...

        class Meta(type):
            @classmethod
            def metaclass_classmethod(metacls) -> str: ...
            @classmethod
            async def metaclass_classmethod2(mcls) -> typing.Any: ...

        class Bacon(dict[str, Any], list[Any], tuple[Any, ...]): ...
        class Ham(dict[str, Incomplete], list[Incomplete], tuple[Incomplete, ...]): ...
        class SpammySpam(dict[str, list[_typeshed.Incomplete]], list[list[typing.Any]]): ...
        class Peanuts(typing_extensions.Any): ...
        class Cashews(Incomplete): ...
        class Peas(Any, Incomplete): ...
        """
    )


@pytest.fixture(scope="session")
def expected_stats_on_example_stub_file() -> AnnotationStats:
    return AnnotationStats(
        annotated_parameters=12,
        unannotated_parameters=5,
        annotated_returns=12,
        unannotated_returns=9,
        explicit_Incomplete_parameters=2,
        explicit_Incomplete_returns=1,
        explicit_Any_parameters=3,
        explicit_Any_returns=2,
        annotated_variables=12,
        explicit_Any_variables=4,
        explicit_Incomplete_variables=2,
        classdefs=9,
        classdefs_with_Any=4,
        classdefs_with_Incomplete=4,
    )


PYRIGHTCONFIG_TEMPLATE = """\
{{
    "typeshedPath": ".",
    // A comment to make this invalid JSON
    "exclude": [
        {}
    ],
}}
"""


@pytest.fixture
def complete_typeshed(
    tmp_path: Path,
    example_stub_source: str,
    real_typeshed_package_names: frozenset[str],
) -> Path:
    typeshed = tmp_path
    for directory in "stdlib", "stubs":
        (typeshed / directory).mkdir()

    (typeshed / "pyrightconfig.json").write_text("{}", encoding="utf-8")
    pyrightconfig_path = typeshed / "pyrightconfig.stricter.json"
    pyrightconfig_path.write_text(PYRIGHTCONFIG_TEMPLATE.format(""), encoding="utf-8")

    for package_name in real_typeshed_package_names:
        package_dir = typeshed / "stubs" / package_name
        package_dir.mkdir()
        write_metadata_text(typeshed, package_name, "version = 0.1")
        source_dir = package_dir / package_name
        source_dir.mkdir()
        for filename in "foo", "bar", "spam", "spammy_spam", "ham", "eggs", "bacon":
            path = source_dir / f"{filename}.pyi"
            path.write_text(example_stub_source, encoding="utf-8")

    functools_path = typeshed / "stdlib" / "functools.pyi"
    functools_path.write_text(example_stub_source, encoding="utf-8")
    return typeshed


@pytest.fixture(scope="session")
def AnnotationStats_fieldnames() -> tuple[str, ...]:
    return tuple(AnnotationStats.__annotations__)


def random_identifier() -> str:
    return "".join(
        random.choice(string.ascii_letters) for _ in range(random.randint(1, 10))
    )


def random_AnnotationStats() -> AnnotationStats:
    return AnnotationStats(*[
        random.randint(0, 1000) for _ in AnnotationStats.__annotations__
    ])


def random_PackageInfo() -> PackageInfo:
    package_name = random_identifier()
    stubtest_strictness = random.choice(list(StubtestStrictness))
    if stubtest_strictness is StubtestStrictness.SKIPPED:
        stubtest_ci_platforms = []
    else:
        stubtest_ci_platforms = [random.choice(["win32", "darwin", "linux"])]
    return PackageInfo(
        package_name=package_name,
        stub_distribution_name=f"types-{package_name}",
        upstream_url=f"https://github.com/who_knows/{package_name}",
        completeness_level=CompletenessLevel.COMPLETE,
        extra_description=None,
        number_of_lines=random.randint(10, 500),
        package_status=random.choice(list(PackageStatus)),
        stubtest_settings=StubtestSettings(
            stubtest_strictness, stubtest_ci_platforms, 55
        ),
        upload_status=random.choice(list(UploadStatus)),
        pyright_setting=random.choice(list(PyrightSetting)),
        annotation_stats=random_AnnotationStats(),
    )


def random_FileInfo() -> FileInfo:
    package_name = random_identifier()
    return FileInfo(
        file_path=Path(
            "stubs", package_name, package_name, f"{random_identifier()}.pyi"
        ),
        parent_package=package_name,
        number_of_lines=random.randint(10, 500),
        pyright_setting=random.choice(list(PyrightSetting)),
        annotation_stats=random_AnnotationStats(),
    )


@pytest.fixture
def random_PackageInfo_sequence() -> Sequence[PackageInfo]:
    return [random_PackageInfo() for _ in range(random.randint(3, 10))]


@pytest.fixture
def random_FileInfo_sequence() -> Sequence[FileInfo]:
    return [random_FileInfo() for _ in range(random.randint(3, 10))]


@pytest.fixture(params=[True, False], ids=["use_string_path", "use_pathlib_path"])
def maybe_stringize_path(
    request: pytest.FixtureRequest,
) -> Callable[[Path], Path | str]:
    def inner(path: Path) -> Path | str:
        if request.param:
            return str(path)
        return path

    inner.__name__ = "maybe_stringize_path"
    return inner



import csv
import io
import json
import os
import re
import sys
import types
import unittest.mock
from collections.abc import Iterator, Sequence
from pathlib import Path
from typing import Any, cast
from unittest import mock

# Make sure not to import rich here, as it's an optional dependency.
# Some tests assert behaviour that's predicated on rich not yet being imported.
import markdown
import pytest
from pytest_mock import MockerFixture
from pytest_subtests import SubTests

import typeshed_stats._cli
import typeshed_stats.gather
from typeshed_stats._cli import SUPPORTED_EXTENSIONS, OutputOption, main
from typeshed_stats.gather import (
    CompletenessLevel,
    PackageInfo,
    PackageStatus,
    PyrightSetting,
    StubtestStrictness,
    UploadStatus,
)

from .conftest import random_package_name

# =========
# Utilities
# =========


@pytest.fixture
def args(typeshed: Path) -> list[str]:
    return ["--typeshed-dir", str(typeshed)]


@pytest.fixture
def disabled_rich() -> Iterator[None]:
    """This is necessary for several of the tests that depend on reading from stdout."""
    assert "rich" not in sys.modules
    with mock.patch.dict("sys.modules", rich=None):
        yield


@pytest.fixture
def mocked_gather_stats_on_multiple_packages(
    random_PackageInfo_sequence: Sequence[PackageInfo], mocker: MockerFixture
) -> None:
    mocker.patch.object(
        typeshed_stats._cli,
        "gather_stats_on_multiple_packages",
        autospec=True,
        return_value=random_PackageInfo_sequence,
    )


def assert_returncode_0(args: list[str]) -> None:
    """Assert the return code is 0 when running `main` with the specified args."""
    with pytest.raises(SystemExit) as exc_info:
        main(args)
    code = exc_info.value.code
    assert code == 0


def assert_argparsing_fails(
    args: list[str],
    *,
    capsys: pytest.CaptureFixture[str] | None = None,
    failure_message: str | None = None,
) -> None:
    """Helper function for lots of argparsing tests."""
    with pytest.raises(SystemExit) as exc_info:
        main(args)
    return_code = exc_info.value.code
    assert return_code == 2
    if failure_message is not None and capsys is not None:
        err = capsys.readouterr().err
        expected_message_found = bool(re.search(failure_message, err))
        assert expected_message_found, err


def _OutputOption_to_argparse(option: OutputOption) -> str:
    return f"--to-{option.name.lower()}"


# ===============================
# Tests for the OutputOption enum
# ===============================


@pytest.mark.parametrize("option_name", [option.name for option in OutputOption])
def test_OutputOption(option_name: str) -> None:
    option = OutputOption[option_name]
    extension = option.file_extension
    assert isinstance(extension, str)
    assert extension in SUPPORTED_EXTENSIONS
    assert extension.startswith(".")
    value_repr, option_repr = repr(option.value), repr(option)
    assert value_repr not in option_repr
    assert option.name in option_repr


def test_OutputOption___repr__() -> None:
    actual_repr = repr(OutputOption.PPRINT)
    assert actual_repr == "OutputOption.PPRINT(extension='.txt')"


def test_OutputOption_from_file_extension() -> None:
    markdown_result = OutputOption.from_file_extension(".md")
    assert markdown_result is OutputOption.MARKDOWN
    pprint_result = OutputOption.from_file_extension(".txt")
    assert pprint_result is OutputOption.PPRINT
    with pytest.raises(ValueError, match="Unsupported file extension"):
        OutputOption.from_file_extension(".xml")


def test_pprinting_conversion(
    random_PackageInfo_sequence: Sequence[PackageInfo],
) -> None:
    # No point in testing the exact type of the output here,
    # it's an implementation detail
    # Just test that it works without raising an exception
    result = OutputOption.PPRINT.convert(random_PackageInfo_sequence)
    assert result is not None


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages")
def test_each_output_option_has_code_written_for_it(
    args: list[str], capsys: pytest.CaptureFixture[str], subtests: SubTests
) -> None:
    options_to_output = {}
    assert_returncode_0(args + ["--pprint"])
    pprint_output = capsys.readouterr().out.strip()
    for option in OutputOption:
        if option is OutputOption.PPRINT:
            continue
        assert_returncode_0(args + [_OutputOption_to_argparse(option)])
        options_to_output[option] = capsys.readouterr().out.strip()

    for option, output in options_to_output.items():
        with subtests.test(option=option.name):
            assert pprint_output not in output
            assert output != pprint_output

    num_different_outputs = len({pprint_output, *options_to_output.values()})
    expected_num_different_outputs = len(OutputOption)
    assert num_different_outputs == expected_num_different_outputs


# ===========================================
# Test that it fails with an unknown argument
# ===========================================


def test_program_fails_with_unknown_argument(args: list[str]) -> None:
    bad_arg = "--foo"
    args.append(bad_arg)
    assert_argparsing_fails(args, failure_message=f"unrecognised arguments: {bad_arg}")


# ========================
# Tests for --typeshed-dir
# ========================


@pytest.mark.parametrize("args", [[], ["foo"], ["--to-csv"]])
def test_argparse_fails_with_no_typeshed_dir_arg(
    capsys: pytest.CaptureFixture[str], args: list[str]
) -> None:
    message = "one of the arguments .+ is required"
    assert_argparsing_fails(args, capsys=capsys, failure_message=message)


def test_invalid_typeshed_dir_arg(
    tmp_path: Path, capsys: pytest.CaptureFixture[str]
) -> None:
    args = ["--typeshed-dir", str(tmp_path)]
    message = "is not a valid typeshed directory"
    assert_argparsing_fails(args, capsys=capsys, failure_message=message)


# ==========================
# Tests for passing packages
# ==========================


@pytest.fixture(params=[1, 2], ids=["one_package", "two_packages"])
def typeshed_with_packages(
    tmp_path: Path, EXAMPLE_PACKAGE_NAME: str, request: pytest.FixtureRequest
) -> Path:
    typeshed = tmp_path
    (typeshed / "stdlib").mkdir()
    stubs_dir = typeshed / "stubs"
    stubs_dir.mkdir()
    (stubs_dir / EXAMPLE_PACKAGE_NAME).mkdir()
    for _ in range(request.param - 1):
        (stubs_dir / random_package_name()).mkdir()
    return typeshed


@pytest.mark.usefixtures("disabled_rich")
class TestPassingPackages:
    _capsys: pytest.CaptureFixture[str]
    _guaranteed_package_name: str

    LOGGING_ARGS = ("--log", "CRITICAL")

    @pytest.fixture(autouse=True)
    def _setup_and_teardown(
        self,
        mocker: MockerFixture,
        EXAMPLE_PACKAGE_NAME: str,
        capsys: pytest.CaptureFixture[str],
    ) -> None:
        patches_to_apply = [
            ("get_package_status", PackageStatus.UP_TO_DATE),
            ("get_completeness_level", CompletenessLevel.COMPLETE),
            (
                "get_upstream_url",
                f"https://github.com/who_knows/{EXAMPLE_PACKAGE_NAME}",
            ),
            ("get_stubtest_strictness", StubtestStrictness.MISSING_STUBS_IGNORED),
            ("get_pyright_setting_for_package", PyrightSetting.STRICT_ON_SOME_FILES),
            ("get_package_extra_description", None),
            ("get_upload_status", UploadStatus.UPLOADED),
            ("get_stubtest_ci_platforms", ["linux"]),
            ("get_stub_distribution_name", "types-foo"),
            ("get_stubtest_allowlist_length", 55),
        ]

        for function_name, return_value in patches_to_apply:
            mocker.patch.object(
                typeshed_stats.gather,
                function_name,
                autospec=True,
                return_value=return_value,
            )

        mocker.patch.dict(
            vars(self), _guaranteed_package_name=EXAMPLE_PACKAGE_NAME, _capsys=capsys
        )

    def _assert_correct_results_printed_to_stdout(
        self, expected_length_of_results: int
    ) -> None:
        stdout = self._capsys.readouterr().out.strip()
        results = eval(stdout, vars(typeshed_stats.gather) | globals())
        assert isinstance(results, dict)
        assert len(results) == expected_length_of_results
        assert all(isinstance(key, str) for key in results)
        assert all(isinstance(value, PackageInfo) for value in results.values())
        guaranteed_package_name = self._guaranteed_package_name
        assert results[guaranteed_package_name].package_name == guaranteed_package_name

    def test_passing_packages_before_typeshed_dir(
        self, typeshed_with_packages: Path
    ) -> None:
        packages_to_pass = os.listdir(typeshed_with_packages / "stubs")
        expected_length_of_results = len(packages_to_pass)
        args = [
            *packages_to_pass,
            "--typeshed-dir",
            str(typeshed_with_packages),
            *self.LOGGING_ARGS,
        ]
        assert_returncode_0(args)
        self._assert_correct_results_printed_to_stdout(expected_length_of_results)

    def test_passing_packages_after_typeshed_dir(
        self, typeshed_with_packages: Path
    ) -> None:
        packages_to_pass = os.listdir(typeshed_with_packages / "stubs")
        expected_length_of_results = len(packages_to_pass)
        args = [
            "--typeshed-dir",
            str(typeshed_with_packages),
            *self.LOGGING_ARGS,
            *packages_to_pass,
        ]
        assert_returncode_0(args)
        self._assert_correct_results_printed_to_stdout(expected_length_of_results)


def test_invalid_packages_given(
    args: list[str], capsys: pytest.CaptureFixture[str]
) -> None:
    message = "'boto' does not have stubs in typeshed!"
    assert_argparsing_fails(args + ["boto"], capsys=capsys, failure_message=message)


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages")
def test_passing_stdlib_as_package(args: list[str]) -> None:
    assert_returncode_0(args + ["stdlib"])


# ===================================
# Tests for --to-file and --overwrite
# ===================================


class ToFileOptionTestsBase:
    _dir: Path
    _args: list[str]
    _capsys: pytest.CaptureFixture[str]

    @pytest.fixture(autouse=True)
    def _setup_and_teardown(
        self,
        tmp_path: Path,
        args: list[str],
        capsys: pytest.CaptureFixture[str],
        mocker: MockerFixture,
    ) -> None:
        mocker.patch.dict(vars(self), _dir=tmp_path, _args=args, _capsys=capsys)


class TestToFileOptionFailureCases(ToFileOptionTestsBase):
    def _assert_fails_with_message(self, message: str) -> None:
        assert_argparsing_fails(
            self._args, capsys=self._capsys, failure_message=message
        )

    @pytest.mark.parametrize(
        ("to_file_args", "failure_message"),
        [
            pytest.param(
                ["--to-file", "bar"],
                "has no file extension",
                id="to_file_with_no_file_extension",
            ),
            pytest.param(
                ["--to-file", "bar.xml"],
                "Unrecognised file extension",
                id="to_file_with_unknown_file_extension",
            ),
            pytest.param(
                ["--to-csv", "--to-file", "bar.csv"],
                "not allowed with argument",
                id="to_file_with_other_output_option",
            ),
        ],
    )
    def test_basic_failure_cases(
        self, to_file_args: list[str], failure_message: str
    ) -> None:
        self._args += to_file_args
        self._assert_fails_with_message(failure_message)

    def test_overwrite_store_true(self, EXAMPLE_PACKAGE_NAME: str) -> None:
        self._args += [
            EXAMPLE_PACKAGE_NAME,
            "--to-file",
            "bar.csv",
            "--overwrite",
            "foo",
        ]
        self._assert_fails_with_message("unrecognized arguments")

    def test_to_file_fails_if_parent_doesnt_exist(self) -> None:
        file_in_fictitious_dir = self._dir / "fiction" / "foo.json"
        self._args += ["--to-file", str(file_in_fictitious_dir)]
        self._assert_fails_with_message("does not exist as a directory!")

    def test_to_file_fails_if_parent_is_not_directory(self) -> None:
        file = self._dir / "file"
        file.write_text("", encoding="utf-8")
        writefile = file / "file2.json"
        self._args += ["--to-file", str(writefile)]
        self._assert_fails_with_message("does not exist as a directory!")

    def test_to_file_fails_if_file_exists(self) -> None:
        # Setup the case where --overwrite=False,
        # and --to-file points to an already existing file
        file_name = "foo.json"
        already_existing_file = self._dir / file_name
        already_existing_file.write_text("", encoding="utf-8")
        self._args += ["--to-file", str(already_existing_file)]
        self._assert_fails_with_message(f'{file_name}" already exists!')


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages")
class TestToFileSuccessCases(ToFileOptionTestsBase):
    def _assert_args_succeed(self) -> None:
        assert_returncode_0(self._args)

    def test_to_file_csv(self) -> None:
        dest = self._dir / "foo.csv"
        self._args += ["--to-file", str(dest)]
        self._assert_args_succeed()
        with dest.open(encoding="utf-8", newline="") as csvfile:
            stats = list(csv.DictReader(csvfile))
        assert all(isinstance(item, dict) for item in stats)
        assert all(isinstance(item["package_name"], str) for item in stats)

    def test_to_file_json(self) -> None:
        dest = self._dir / "foo.json"
        self._args += ["--to-file", str(dest)]
        self._assert_args_succeed()
        with dest.open(encoding="utf-8") as jsonfile:
            result = json.load(jsonfile)
        assert isinstance(result, list)
        assert all(isinstance(item, dict) for item in result)
        assert all(isinstance(item["package_name"], str) for item in result)

    def test_to_file_markdown(self) -> None:
        dest = self._dir / "foo.md"
        self._args += ["--to-file", str(dest)]
        self._assert_args_succeed()
        with dest.open(encoding="utf-8") as markdown_file:
            markdown.markdown(markdown_file.read())

    def test_to_file_txt(self) -> None:
        dest = self._dir / "foo.txt"
        self._args += ["--to-file", str(dest)]
        self._assert_args_succeed()
        with dest.open(encoding="utf-8") as txtfile:
            source = txtfile.read()
        results = eval(source, vars(typeshed_stats.gather) | globals())
        assert isinstance(results, dict)
        assert all(isinstance(key, str) for key in results)
        assert all(isinstance(value, PackageInfo) for value in results.values())

    def test_overwrite_argument(self) -> None:
        # Setup the case where --overwrite=True,
        # and --to-file points to an already existing file
        already_existing_file = self._dir / "foo.json"
        already_existing_file.write_text("", encoding="utf-8")
        self._args += ["--to-file", str(already_existing_file), "--overwrite"]
        self._assert_args_succeed()


# ================================
# Tests for various output options
# ================================


@pytest.fixture
def mocked_pprint_dot_pprint(mocker: MockerFixture) -> mock.MagicMock:
    return mocker.patch("pprint.pprint", autospec=True)


class OutputOptionsPrintingToTerminalTestsBase:
    _args: list[str]
    _capsys: pytest.CaptureFixture[str]

    @pytest.fixture(autouse=True)
    def _setup_and_teardown(
        self, args: list[str], capsys: pytest.CaptureFixture[str], mocker: MockerFixture
    ) -> None:
        mocker.patch.dict(
            vars(self), _args=(args + ["--log", "CRITICAL"]), _capsys=capsys
        )


class TestOutputOptionsToTerminalFailureCases(OutputOptionsPrintingToTerminalTestsBase):
    def _assert_fails_with_message(self, message: str) -> None:
        assert_argparsing_fails(self._args, failure_message=message)

    @pytest.mark.parametrize(
        "option",
        [
            (
                "--pprint"
                if option is OutputOption.PPRINT
                else _OutputOption_to_argparse(option)
            )
            for option in OutputOption
        ],
    )
    def test_bad_use_of_output_option(
        self, option: str, EXAMPLE_PACKAGE_NAME: str
    ) -> None:
        bogus_argument = "foo"
        self._args += [EXAMPLE_PACKAGE_NAME, option, bogus_argument]
        self._assert_fails_with_message(f"unrecognized arguments: {bogus_argument}")

    @pytest.mark.parametrize(
        "options",
        [
            ["--pprint", "--to-csv"],
            ["--to-csv", "--to-markdown"],
            ["--to-csv", "--to-json", "--to-markdown"],
        ],
    )
    def test_passing_multiple_options_fails(self, options: list[str]) -> None:
        self._args += options
        self._assert_fails_with_message("not allowed with argument")


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages", "disabled_rich")
class TestOutputOptionsToTerminalSuccessCases(OutputOptionsPrintingToTerminalTestsBase):
    def _assert_outputoption_works(self, option: str) -> None:
        args = self._args + [option]
        assert_returncode_0(args)

    def _get_stdout(self) -> str:
        return self._capsys.readouterr().out.strip()

    def test_to_json(self) -> None:
        self._assert_outputoption_works("--to-json")
        result = json.loads(self._get_stdout())
        assert isinstance(result, list)
        assert all(isinstance(item, dict) for item in result)
        assert all(isinstance(item["package_name"], str) for item in result)

    def test_to_csv(self) -> None:
        self._assert_outputoption_works("--to-csv")
        csvfile = io.StringIO(self._get_stdout(), newline="")
        stats = list(csv.DictReader(csvfile))
        assert all(isinstance(item, dict) for item in stats)
        assert all(isinstance(item["package_name"], str) for item in stats)

    def test_to_markdown(self) -> None:
        self._assert_outputoption_works("--to-markdown")
        result = self._get_stdout()
        markdown.markdown(result)

    def test_pprint_option_with_rich_available(self) -> None:
        sys_modules = cast(dict[str, types.ModuleType | None], sys.modules)
        assert sys_modules["rich"] is None
        del sys.modules["rich"]

        import rich

        old_rich_print = rich.print
        rich_print_called = False

        def new_rich_print(*args: Any, **kwargs: Any) -> None:
            nonlocal rich_print_called
            rich_print_called = True
            old_rich_print(*args, **kwargs)

        rich.print = new_rich_print

        self._assert_outputoption_works("--pprint")
        assert rich_print_called

        rich.print = old_rich_print
        sys_modules["rich"] = None

    @pytest.mark.usefixtures("disabled_rich")
    def test_pprint_option_with_rich_unavailable(
        self, mocked_pprint_dot_pprint: mock.MagicMock
    ) -> None:
        self._assert_outputoption_works("--pprint")
        mocked_pprint_dot_pprint.assert_called_once()

    @pytest.mark.usefixtures("disabled_rich")
    @pytest.mark.parametrize(
        "option",
        [
            _OutputOption_to_argparse(option)
            for option in OutputOption
            if option is not OutputOption.PPRINT
        ],
    )
    def test_other_options_with_rich_unavailable(
        self, option: str, mocked_pprint_dot_pprint: mock.MagicMock
    ) -> None:
        self._assert_outputoption_works(option)
        mocked_pprint_dot_pprint.assert_not_called()


# ========================
# Tests for --log argument
# ========================


@pytest.mark.parametrize("bad_arg", ["FOO", "getLogger"])
def test_bad_log_argument(args: list[str], bad_arg: str) -> None:
    args += ["--log", bad_arg]
    assert_argparsing_fails(
        args, failure_message=f"invalid choice: {bad_arg!r} (choose from"
    )


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages")
@pytest.mark.parametrize(
    ("logging_level", "logging_expected"),
    [
        pytest.param("NOTSET", False, id="NOTSET_logging_not_expected"),
        pytest.param("DEBUG", True, id="DEBUG_logging_expected"),
        pytest.param("INFO", True, id="INFO_logging_expected"),
        pytest.param("WARNING", False, id="WARNING_logging_not_expected"),
        pytest.param("ERROR", False, id="ERROR_logging_not_expected"),
        pytest.param("CRITICAL", False, id="CRITICAL_logging_not_expected"),
    ],
)
def test_logs_to_terminal_with_info_or_lower(
    args: list[str],
    logging_level: str,
    logging_expected: bool,
    caplog: pytest.LogCaptureFixture,
) -> None:
    args += ["--log", logging_level]
    assert_returncode_0(args)
    logging_occurred = bool(caplog.text.strip())
    assert logging_occurred is logging_expected


# ================================
# Tests for `--upstream-repo` flag
# ================================


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages")
def test_upstream_url_flag(
    typeshed: Path, args: list[str], random_PackageInfo_sequence: Sequence[PackageInfo]
) -> None:
    random_PackageInfo_sequence[-1].upstream_url = None

    package_names = [package.package_name for package in random_PackageInfo_sequence]
    for name in package_names:
        (typeshed / "stubs" / name).mkdir(exist_ok=True)

    args += package_names
    args += ["--upstream-repo"]

    webbrowser_mock_path = "webbrowser.open_new_tab"
    with unittest.mock.patch(webbrowser_mock_path, autospec=True) as webbrowser_mock:
        assert_returncode_0(args)
        opened_urls = {call.args[0] for call in webbrowser_mock.call_args_list}
        assert len(opened_urls) == len(random_PackageInfo_sequence) - 1
        expected_urls = {
            package.upstream_url for package in random_PackageInfo_sequence[:-1]
        }
        assert expected_urls == opened_urls


@pytest.mark.usefixtures("mocked_gather_stats_on_multiple_packages")
def test_upstream_url_flag_no_known_urls(
    typeshed: Path, args: list[str], random_PackageInfo_sequence: Sequence[PackageInfo]
) -> None:
    for package in random_PackageInfo_sequence:
        package.upstream_url = None

    package_names = [package.package_name for package in random_PackageInfo_sequence]
    for name in package_names:
        (typeshed / "stubs" / name).mkdir(exist_ok=True)

    assert_argparsing_fails(args + package_names + ["--upstream-repo"])


# ====================================
# Tests for KeyboardInterrupt-handling
# ====================================


def test_KeyboardInterrupt_caught(
    args: list[str], capsys: pytest.CaptureFixture[str]
) -> None:
    args += ["--log", "CRITICAL"]
    with (
        pytest.raises(SystemExit) as exc_info,
        mock.patch.object(
            typeshed_stats.gather,
            "gather_stats_on_package",
            autospec=True,
            side_effect=KeyboardInterrupt(),
        ),
    ):
        main(args)

    stderr = capsys.readouterr().err
    num_stderr_lines = len(stderr.strip().splitlines())
    assert num_stderr_lines == 1
    assert "Interrupted!" in stderr
    return_code = exc_info.value.code
    assert return_code == 2


# ========================================
# Integration test for --download-typeshed
# ========================================


@pytest.mark.requires_network
def test_integration_with_download_typeshed(caplog: pytest.LogCaptureFixture) -> None:
    assert_returncode_0(["--download-typeshed"])
    assert "Cloning typeshed" in caplog.text



"""Convenience script to run all checks locally."""

import argparse
import subprocess
import sys
from pathlib import Path
from typing import Literal, overload

FILES_TO_CHECK = ("src", "tests", "scripts", "website_macros.py")


@overload
def run_checks(
    *,
    regenerate_examples: Literal[True],
    typeshed_dir: Path,
    download_typeshed: Literal[False] = ...,
) -> None: ...


@overload
def run_checks(
    *,
    regenerate_examples: Literal[True],
    download_typeshed: Literal[True],
    typeshed_dir: None = ...,
) -> None: ...


@overload
def run_checks(
    *,
    regenerate_examples: Literal[False],
    download_typeshed: Literal[False] = ...,
    typeshed_dir: None = ...,
) -> None: ...


@overload
def run_checks(
    *,
    regenerate_examples: bool = ...,
    typeshed_dir: Path | None = ...,
    download_typeshed: bool = ...,
) -> None: ...


def run_checks(
    *,
    regenerate_examples: bool = False,
    typeshed_dir: Path | None = None,
    download_typeshed: bool = False,
) -> None:
    """Run the checks."""
    print("\nRunning ruff...")
    subprocess.run(["ruff", "check", "."], check=False)

    print("\nRunning ruff formatter...")
    ruff_format_result = subprocess.run(["ruff", "format", "."], check=False)
    if ruff_format_result.returncode == 123:
        print("Exiting early since `ruff format` failed!")
        raise SystemExit(1)

    print("\nRunning mypy...")
    subprocess.run(["mypy"], check=True)

    print("\nRunning pytest...")
    for command in "coverage run -m pytest", "coverage report":
        subprocess.run(command.split(), check=True)

    if regenerate_examples:
        args = [sys.executable, "regenerate_examples.py"]
        if download_typeshed:
            args.append("--download-typeshed")
        else:
            assert typeshed_dir is not None
            args.extend(["--typeshed-dir", str(typeshed_dir)])
        subprocess.run(args, check=True)


def main() -> None:
    """Parse command-line args, run the checks."""
    package_dir = Path("src", "typeshed_stats")
    running_from_root = package_dir.exists() and package_dir.is_dir()
    assert running_from_root, "Must run this script from the repository root!"

    parser = argparse.ArgumentParser("Script to regenerate examples")
    parser.add_argument("-r", "--regen-examples", action="store_true")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-t", "--typeshed-dir", type=Path)
    group.add_argument("-d", "--download-typeshed", action="store_true")
    args = parser.parse_args()
    if args.regen_examples:
        if not args.typeshed_dir and not args.download_typeshed:
            parser.error(
                "Must specify --typeshed-dir on the command line"
                " if --regen-examples is specified"
            )
    elif args.typeshed_dir:
        parser.error(
            "--typeshed-dir has no meaning if --regen-examples is not specified"
        )
    run_checks(
        regenerate_examples=args.regen_examples,
        typeshed_dir=args.typeshed_dir,
        download_typeshed=args.download_typeshed,
    )


if __name__ == "__main__":
    main()
    raise SystemExit(0)



# pyright: reportUnusedFunction=false
"""Macros and variables for the mkdocs-macros plugin."""

import builtins
import csv
import datetime as dt
import shutil
import subprocess
import types
from abc import abstractmethod
from collections.abc import Callable
from enum import Enum
from pathlib import Path
from typing import Any, Protocol, TypeGuard, TypeVar, get_args, get_origin

import attrs
import markdown

import typeshed_stats.gather
import typeshed_stats.serialize

CF = TypeVar("CF", bound=Callable[..., Any])


class Env(Protocol):
    """Protocol representing the `mkdocs_macros.MacrosPlugin` object."""

    variables: dict[str, Any]
    conf: dict[str, Any]

    @abstractmethod
    def macro(self, /, func: CF) -> CF:
        """Register a function as a macro."""


def define_env(env: Env) -> None:
    """Define environment variables."""
    # Variables needed for cli.md
    help_result = subprocess.run(
        ["typeshed-stats", "--help"], text=True, capture_output=True, check=True
    )
    env.variables["cli_help"] = help_result.stdout + "\n"

    # Variables needed for serialize.md
    env.variables["serialize"] = typeshed_stats.serialize

    # Variables needed for gather.md
    @env.macro
    def is_enum(x: object) -> TypeGuard[type[Enum]]:
        return isinstance(x, type) and issubclass(x, Enum)

    @env.macro
    def get_field_description(typ: object) -> str:
        """Get a description of the type of a field in an `attrs` class.

        A helper function for the `gather.md.jinja` template.
        """
        if isinstance(typ, types.UnionType):
            return r" \| ".join(map(get_field_description, get_args(typ)))
        if isinstance(typ, types.GenericAlias):
            return (
                get_field_description(get_origin(typ))
                + "["
                + ", ".join(map(get_field_description, get_args(typ)))
                + "]"
            )
        if isinstance(typ, type):
            typ_name = typ.__name__
            if typ_name in typeshed_stats.gather.__all__:
                return f"[`{typ_name}`][typeshed_stats.gather.{typ_name}]"
            if typ_name in dir(builtins):
                return f"[`{typ_name}`][{typ_name}]"
            if typ is Path:
                return "[`Path`][pathlib.Path]"
            if typ is types.NoneType:
                return "[`None`][None]"
            return f"`{typ_name}`"
        return f"`{typ}`"

    env.variables.update({"gather": typeshed_stats.gather, "attrs": attrs})

    # Variables needed for stats.md and stats-csv.md
    stats_as_csv: list[dict[str, str | int]] = []
    with Path("examples", "example.csv").open(newline="", encoding="utf-8") as csvfile:
        for line in csv.DictReader(csvfile):
            row: dict[str, str | int] = {}
            for key, val in line.items():
                if val.isdigit():
                    row[key] = int(val)
                elif key == "upstream_url":
                    row[key] = "(unknown)" if val == "-" else f"[{val}]({val})"
                else:
                    row[key] = val
            stats_as_csv.append(row)

    @env.macro
    def is_int(x: object) -> TypeGuard[int]:
        return isinstance(x, int)

    env.variables.update({
        "last_update_time": dt.datetime.now(dt.UTC).strftime("%H:%M UTC on %Y-%m-%d"),
        "num_packages": len(stats_as_csv),
        "formatted_stats": Path("examples", "example.md").read_text(encoding="utf-8"),
        "num_lines": sum(int(s["number_of_lines"]) for s in stats_as_csv),
        "stats_as_csv": stats_as_csv,
        "markdown": markdown,
    })


def on_post_build(env: Env) -> None:
    """Define post-build actions."""
    shutil.copyfile(
        Path("examples", "example.csv"), Path(env.conf["site_dir"], "stats_as_csv.csv")
    )



"""Test __all__ in __init__.py and all submodules."""

import builtins
import importlib
import pkgutil
import sys
import types
from typing import Final
from unittest import mock

# Make sure not to import rich or markdown here, since they're optional dependencies
# Some tests assert behaviour that's predicated on these modules not yet being imported
import pytest

import typeshed_stats

ALL_SUBMODULES: Final = [
    importlib.import_module(f"typeshed_stats.{m.name}")
    for m in pkgutil.iter_modules(typeshed_stats.__path__)
    if m.name != "_version"
]


@pytest.fixture(params=ALL_SUBMODULES)
def submodule(request: pytest.FixtureRequest) -> types.ModuleType:
    return request.param  # type: ignore[no-any-return]


def test_submodule_has___all__(submodule: types.ModuleType) -> None:
    assert hasattr(submodule, "__all__")


def test_submodule__all___is_valid(submodule: types.ModuleType) -> None:
    assert isinstance(submodule.__all__, list)
    assert all(isinstance(item, str) for item in submodule.__all__)


def test_all_public_names_in___all__(submodule: types.ModuleType) -> None:
    """Test that all names not in `__all__` are marked as private."""
    submodule_name = submodule.__name__

    def is_from_other_module(obj: object) -> bool:
        return getattr(obj, "__module__", submodule_name) != submodule_name

    def is_private_or_imported_symbol(name: str, value: object) -> bool:
        return (
            name.startswith("_")
            or isinstance(value, types.ModuleType)
            or is_from_other_module(value)
            or name in vars(builtins)
        )

    public_names_not_in___all__ = {
        name
        for name, value in vars(submodule).items()
        if (
            name not in submodule.__all__
            and not is_private_or_imported_symbol(name, value)
        )
    }

    assert not public_names_not_in___all__, (
        f"Public names not in __all__: {public_names_not_in___all__!r}"
    )


# ========================================
# Test the package doesn't import on <3.11
# ========================================


@mock.patch.object(sys, "version_info", new=(3, 9, 8, "final", 0))
@pytest.mark.parametrize("module_name", [mod.__name__ for mod in ALL_SUBMODULES])
def test_import_fails_on_less_than_3_point_11(module_name: str) -> None:
    for submod in ALL_SUBMODULES:
        submod_name = submod.__name__
        if submod_name in sys.modules:
            del sys.modules[submod_name]
    with pytest.raises(ImportError, match=r"Python 3\.11\+ is required"):
        importlib.import_module(module_name)
